(******************************************************************************)
(* generator (stack machine)                                                  *)
(* (C) 2010..2021 Steffen Solyga <solyga@gmx.de>                              *)
(******************************************************************************)
IMPLEMENTATION MODULE Gen;

IMPORT FIO;
IMPORT Comp;

CONST
  debug = FALSE;
  out ::= FIO.StandardOutput;
  err ::= FIO.ErrorOutput;
  dbg ::= FIO.ErrorOutput;

  ocMax = 32;

TYPE
  TOpdType = ( tVal, tIdx );
  TOpdDesc = RECORD
    CASE t: TOpdType OF
    | tVal: v: Comp.TValue;
    | tIdx: i: Comp.TIndex;
    END;
  END;
  TIdx = [ 1 .. ocMax ];

  TStr = ARRAY [ 0 .. 2 ] OF CHAR;

VAR
  opds: ARRAY TIdx OF TOpdDesc; (* (compile time) stack *)
  oc: CARDINAL;                 (* operand count *)
  vc: CARDINAL;                 (* variable count (= base index of temp vars) *)
  as: CARDINAL;  (* acc state (as left by Sto & Opr), 0: free, TIdx: result *)
  bc: CARDINAL;  (* backup count *)

  opns: ARRAY TOperation OF TStr;
  first: BOOLEAN;


PROCEDURE DbgStk();
VAR i: CARDINAL;
BEGIN
  i := oc;
  WHILE i > 0 DO
    FIO.WrStr( dbg, 'opds[' );
    FIO.WrCard( dbg, i, 0 );
    FIO.WrStr( dbg, '] = ' );
    FIO.WrCard( dbg, ORD(opds[i].t), 0 );
    FIO.WrStr( dbg, ' ' );
    CASE opds[ i ].t OF
    | tVal: FIO.WrInt( dbg, opds[i].v, 11 );
    | tIdx: FIO.WrCard( dbg, opds[i].i, 2 );
    END;
    DEC( i );
    FIO.WrLn( dbg );
  END;
END DbgStk;

PROCEDURE DbgH();
BEGIN
  FIO.WrStr( dbg, 'Gen.Debug: vc oc bc as' );
  FIO.WrLn( dbg );
END DbgH;

PROCEDURE DbgS( s: ARRAY OF CHAR );
BEGIN
  IF first THEN DbgH(); first := FALSE END;
  FIO.WrStr( dbg, 'Gen.Debug: ' );
  FIO.WrCard( dbg, vc, 2 );
  FIO.WrStr( dbg, '  ' );
  FIO.WrCard( dbg, oc, 1 );
  FIO.WrStr( dbg, ' ' );
  FIO.WrCard( dbg, bc, 2 );
  FIO.WrStr( dbg, ' ' );
  FIO.WrCard( dbg, as, 2 );
  FIO.WrStr( dbg, ':  ' );
  FIO.WrStr( dbg, s );
END DbgS;

PROCEDURE DbgX();
BEGIN
  FIO.WrLn( dbg );
END DbgX;

PROCEDURE DbgV( v: TValue );
BEGIN
  FIO.WrStr( dbg, '  ' );
  FIO.WrInt( dbg, v, 0 );
  FIO.WrLn( dbg );
END DbgV;

PROCEDURE DbgO( o: TOffset );
BEGIN
  FIO.WrStr( dbg, '   ' );
  FIO.WrCard( dbg, o, 1 );
  FIO.WrLn( dbg );
END DbgO;


PROCEDURE Err( s: ARRAY OF CHAR );
BEGIN
  FIO.WrStr( err, 'Gen: ' );
  FIO.WrStr( err, s );
  FIO.WrStr( err, '.' );
  FIO.WrLn( err );
  HALT;
END Err;

PROCEDURE Value( v: TValue ): Comp.TValue;
BEGIN
  RETURN v;
END Value;

PROCEDURE Index( o: TOffset ): Comp.TIndex;
BEGIN
  IF o > MAX( Comp.TIndex ) THEN Err( 'Invalid offset' ) END;
  RETURN o;
END Index;

PROCEDURE ChkPush();
BEGIN
  IF oc >= ocMax THEN Err( 'Stack overflow' ) END;
END ChkPush;

PROCEDURE ChkPop( n: CARDINAL );
BEGIN
  IF oc < n THEN Err( 'Stack underflow' ) END;
END ChkPop;


PROCEDURE Val( v: TValue );
BEGIN
  IF debug THEN DbgS( 'Val' ); DbgV( v ) END;
  ChkPush(); INC( oc );
  opds[ oc ].t := tVal;
  opds[ oc ].v := Value( v );
END Val;

PROCEDURE Lod( o: TOffset );
BEGIN
  IF debug THEN DbgS( 'Lod' ); DbgO( o ) END;
  ChkPush(); INC( oc );
  opds[ oc ].t := tIdx;
  opds[ oc ].i := Index( o );
END Lod;

PROCEDURE Bak( i: TIdx ): BOOLEAN;
BEGIN
  RETURN ( opds[i].t = tIdx ) & ( opds[i].i >= vc );
END Bak;

PROCEDURE Lda( i: TIdx );
BEGIN
  IF as = i THEN RETURN END;  (* result already in acc *)
  IF as > 0 THEN  (* save *)
    IF ~ Bak( as ) THEN  (* allocate *)
      IF vc + bc > MAX( Comp.TIndex ) THEN Err( 'Out of backup space' ) END;
      opds[ as ].t := tIdx;
      opds[ as ].i := vc + bc;
      INC( bc );
    END;
    Comp.Sto( opds[as].i );
  END;
  CASE opds[ i ].t OF
  | tVal: Comp.Ldv( opds[i].v );
  | tIdx: Comp.Ldi( opds[i].i );
  END;
  as := 0;  (* no result *)
END Lda;

PROCEDURE Dec();
BEGIN
  IF Bak( oc ) THEN DEC( bc ) END;  (* garbage collector *)
  DEC( oc );
END Dec;

PROCEDURE Sto( o: TOffset );
BEGIN
  IF debug THEN DbgS( 'Sto' ); DbgO( o ) END;
  ChkPop( 1 ); Lda( oc ); Comp.Sto( Index(o) ); Dec(); as := 0;
END Sto;

PROCEDURE Calc( o: Comp.TOperation );
BEGIN
  ChkPop( 2 ); Lda( oc - 1 );
  CASE opds[ oc ].t OF
  | tVal: Comp.Opv( o, opds[oc].v );
  | tIdx: Comp.Opi( o, opds[oc].i );
  END;
  Dec(); as := oc;
END Calc;

PROCEDURE Opr( o: TOperation );
BEGIN
  IF debug THEN DbgS( opns[o] ); DbgX() END;
  CASE o OF
  | oWrt: Lda( oc ); Comp.Out(); Dec(); as := 0;
  | oNeg: Lda( oc ); Comp.Neg(); as := oc;
  | oAdd: Calc( Comp.oAdd ); 
  | oSub: Calc( Comp.oSub ); 
  | oMul: Calc( Comp.oMul );
  | oDiv: Calc( Comp.oDiv );
  | oNew: IF oc > 0 THEN Err( 'New var while stack not empty' ) END;
          INC( vc ); IF vc > Comp.cnt THEN Err( 'Too many variables' ) END;
  ELSE
    Err( 'Bug: Invalid operation' );
  END;
END Opr;

BEGIN
  oc := 0;
  vc := 0;
  as := 0;
  bc := 0;
  opns[ oNeg ] := 'Neg';
  opns[ oWrt ] := 'Wrt';
  opns[ oAdd ] := 'Add';
  opns[ oSub ] := 'Sub';
  opns[ oMul ] := 'Mul';
  opns[ oDiv ] := 'Div';
  opns[ oNew ] := 'New';
  first := TRUE;
END Gen.
